import objects
import system
from global_defines import Lambda
from global_defines import data
import types
import globaltext
#import time

__illegal_keywords = ["excess", "modifier"]
__dont_use_excess = ["hitpoints", "maxhitpoints", "armor_display_value"]
__permanent_bonus = ["hitpoints"]

BONUS = 0
REMOVE = 1
TIMEOUT = 2
CALLBACK = 3
REVERSED = 4
COUNT = 5

debug = False

__doc__ = """This module handles all temporary bonuses/maluses for player chars.

Bonuses can be assigned to skills, attributes and hitpoints.
The bonus system ensures that skills and attributes don't go below one because of maluses.
No such protection exists for hitpoints. A malus to hitpoints or maxhitpoints might kill
the char.
All bonuses/maluses are temporary and can be removed via remove_on_leave, timeout or a call
to remove_temp_bonus.
The sole exception is hitpoints. Every bonus/malus to hitpoints is final and must be healed
or removed/hurt by normal means. Even with bonuses hitpoints can't go higher than
maxhitpoints.
"""

def __sanity_check(char):
	"""helper function, verifies after each add/reverse/remove that bonus system works correct"""
	if debug:
		import skill_tables
		properties = skill_tables.attributes + skill_tables.skills + ["hitpoints", "maxhitpoints", "armor_display_value"]
	
		bonus_table = {}
		if objects.has_attribute(char, "bonus_table"):
			bonus_table = objects.get_attribute(char, "bonus_table")
	
		for property in bonus_table:
			assert property in properties
	
			value = objects.get_attribute(char, property)
			if type(value) not in (types.IntType, types.LongType) and property not in ["hitpoints", "maxhitpoints", "armor_display_value"]:
				exc = "illegal type %s in attribute/skill '%s' for char '%s' (value: <%s>)" % (
						str(type(value)), property, char, repr(value))
				raise TypeError, exc 
			assert value >= 1 or property in __dont_use_excess
	
			modifier = 0
			bonus_sum = 0
	
			# make sure necessary keys are in the table_entry
			table_entry = bonus_table[property]
			keys = table_entry.keys()
			needed = __illegal_keywords
			assert filter(lambda elem: elem in keys, needed), "%s isn't in %s" % (str(needed), str(keys))
	
			for key in table_entry:
				if key == "modifier":
					modifier = table_entry[key]
					if property in skill_tables.attributes:
						assert value - modifier <= 20, "%s - %s" % (str(property), str(table_entry))
					if property in skill_tables.skills:
						assert value - modifier <= 99, "%s - %s" % (str(property), str(table_entry))
				elif key == "excess":
					bonus_sum += table_entry[key]
				else:
					bonus_sum += table_entry[key][BONUS]
			assert bonus_sum==modifier, "%s != %s" % (str(property), str(table_entry))
			
			properties.remove(property)
	
		for property in properties:
			value = objects.get_attribute(char, property)
			if property not in __dont_use_excess and type(value) not in (types.IntType, types.LongType):
				exc = "illegal type %s in attribute/skill '%s' for char '%s' (value: <%s>)" % (
						str(type(value)), property, char, repr(value))
				raise TypeError, exc 
			if property in skill_tables.attributes:
				assert value >= 1, "value to low for %s: %s" % (str(property), repr(value))
				assert value <= 20, "attribute %s to high: %s" % (str(property), str(value))
			if property in skill_tables.skills:
				assert value >= 1, "value to low for %s: %s" % (str(property), repr(value))
				assert value <= 99, "skill %s to high: %s" % (str(property), str(value))

def __modify_temp_bonus(char, property, bonus, keyword, timeout=None, callback=None, remove_on_leave=False):
	"""helper function for add_... reverse_... and remove_..."""

	assert keyword not in __illegal_keywords, "illegal keyword, not allowed keywords are: %s" % str(__illegal_keywords)
	assert not callback or timeout, "callback is set but timeout isn't"

	# prepare table
	bonus_table = {}
	if objects.has_attribute(char, "bonus_table"):
		bonus_table = objects.get_attribute(char, "bonus_table")
	
	# prepare table entry for this property
	if property in bonus_table:
		table_entry = bonus_table[property]
	else:
		table_entry = {}
		table_entry["excess"] = 0
		table_entry["modifier"] = 0
		bonus_table[property] = table_entry
	
	# get current property value
	value = objects.get_attribute(char, property)
	
	# calc new_value
	if type(bonus)==int:
		new_value = value + bonus
	elif type(bonus)==float:
		new_value = value * bonus
		new_value = int(round(new_value))
		bonus = int(new_value - value)
	else: assert False
	if not keyword in table_entry:
		table_entry[keyword] = [None]*COUNT
		table_entry[keyword][BONUS] = 0
		table_entry[keyword][REMOVE] = False
		table_entry[keyword][REVERSED] = False
	if property not in __permanent_bonus:
		# set new bonus (possibly replacing an existing bonus with the same keyword)
		new_value -= table_entry[keyword][BONUS]
		table_entry["modifier"] -= table_entry[keyword][BONUS]
		table_entry[keyword][BONUS] = bonus
		table_entry["modifier"] += bonus

	if property not in __dont_use_excess:
		# make sure new value is >= 1
		while new_value < 1:
			new_value += 1
			table_entry["excess"] += 1
			table_entry["modifier"] += 1
	
		# make sure excess is as low as possible, it's only meant to keep the value >= 1
		while new_value > 1 and table_entry["excess"] > 0:
			new_value -= 1
			table_entry["excess"] -= 1
			table_entry["modifier"] -= 1
	
	# register timeout and start timer (if it isn't running already)
	if timeout:
		current_time = system.get_mission_time()
		timeout_time = current_time + timeout
		if not callback:
			callback = Lambda("remove_temp_bonus('%s', '%s')" % (char, keyword))
		table_entry[keyword][TIMEOUT] = str(timeout_time)
		table_entry[keyword][CALLBACK] = callback
		
		if not system.is_global_timer_active("bonus_malus_timer"):
			system.notify_global_timer("bonus_malus_timer", 15*1000, repeating=True)
		
		if not hasattr(data("bonus_malus"), "timeout_chars"):
			data("bonus_malus").timeout_chars = []
		if char not in data("bonus_malus").timeout_chars:
			data("bonus_malus").timeout_chars.append(char)
	else:
		table_entry[keyword][TIMEOUT] = None
		table_entry[keyword][CALLBACK] = None
	
	if remove_on_leave:
		table_entry[keyword][REMOVE] = True
	else:
		table_entry[keyword][REMOVE] = False
	
	if debug:
		print "changed %s from %d to %d. bonus %d" % (property, value, new_value, bonus)
		print table_entry
	
	# save data
	objects.set_attribute(char, "bonus_table", bonus_table)
	# print "ZZZ> ", char, property, new_value
	objects.set_attribute(char, property, new_value)
	# print "ZZZ< ", char, property, new_value	
	
	if char in system.get_pcs():
		system.update_ui()
	
	return bonus

def bonus_malus_timer():
	"""Called by system_events.on_timer to check for bonus timeouts.
	
Is called every 15 seconds."""
	if not hasattr(data("bonus_malus"), "timeout_chars"):
		data("bonus_malus").timeout_chars = []
	
	callbacks = []
	remove_chars = []
	current_time = system.get_mission_time()
	for char in data("bonus_malus").timeout_chars:
		# prepare table
		bonus_table = {}
		if objects.has_attribute(char, "bonus_table"):
			bonus_table = objects.get_attribute(char, "bonus_table")
		
		remove_char_from_timeout_list = True
		for property in bonus_table:
			table_entry = bonus_table[property]

			for keyword in table_entry:
				if not keyword in __illegal_keywords:
					if table_entry[keyword][TIMEOUT]:
						if current_time > table_entry[keyword][TIMEOUT]:
							callbacks.append(table_entry[keyword][CALLBACK])
							table_entry[keyword][TIMEOUT] = None
							table_entry[keyword][CALLBACK] = None
						else:
							remove_char_from_timeout_list = False

		if remove_char_from_timeout_list:
			remove_chars.append(char)
	
	for char in remove_chars:
		data("bonus_malus").timeout_chars.remove(char)
		
	for callback in callbacks:
		callback()	# execute callback

def add_temp_bonuses(char, property_table, keyword, timeout=None, callback=None, remove_on_leave=False):
	"""Adds a temporary bonuses to severale attributes/skills of the char.
	
Parameters:
	char     - Id of the character
	property_table - dict with property/bonus pairs
	keyword  - Keyword like "increase moral effect" or "drug effect".
	           Is used to identify this change and to remove it with remove_temp_bonus.
	           Changes to different properties can use the same keyword, i.e.
	           "drug effect" could be applied to strength, agility and charisma and all
	           All bonuses with the same keyword are removed at once with remove_temp_bonus.
	timeout  - Time string like "12:00:00"
	           None if the bonus has no timeout
	callback - Function to call when timeout is reached
	           If this is None remove_temp_bonus is called.
	           callback is ignored if timeout is None.
	           callback will only be called once, not for each property set.
	remove_on_leave - If this is True the bonus is removed, when the char leaves the
	           player party.
"""
	for property in property_table.keys():
		__modify_temp_bonus(char, property, property_table[property], keyword, timeout, callback, remove_on_leave)
		if timeout or callback: # only one callback wanted
			timeout = None
			callback = None
	
	__sanity_check(char)

def add_temp_bonus(char, property, bonus, keyword, timeout=None, callback=None, remove_on_leave=False):
	"""Adds a temporary bonus to a characters attribute or skill.
	
Parameters:
	char     - Id of the character
	property - One of the attributes named in skill_tables.attributes or
	           one of the skill named in skill_tables.skills or "hitpoints" ot "maxhitpoints"
	bonus    - Bonus or malus applied to property.
	           If it is an integer it will simply be added to property.
	           If it is a float it will be multiplied with property.
	keyword  - Keyword like "increase moral effect" or "drug effect".
	           Is used to identify this change and to remove it with remove_temp_bonus.
	           Changes to different properties can use the same keyword, i.e.
	           "drug effect" could be applied to strength, agility and charisma and all
	           All bonuses with the same keyword are removed at once with remove_temp_bonus.
	timeout  - Time string like "12:00:00"
	           None if the bonus has no timeout
	callback - Function to call when timeout is reached
	           If this is None remove_temp_bonus is called.
	           callback is ignored if timeout is None.
	remove_on_leave - If this is True the bonus is removed, when the char leaves the
	           player party.

Returns:
	(integer) the absolute bonus or malus (mostly needed for factor bonuses
	i.e. bonus 1.2 on a property of 70 returns 14)
"""
	
	bonus = __modify_temp_bonus(char, property, bonus, keyword, timeout, callback, remove_on_leave)
	
	__sanity_check(char)
	
	return bonus


def remove_temp_bonus(char, keyword):
	"""Removes all bonuses with the right keyword from char.
	
No callbacks are fired by this."""
	
	if not objects.has_attribute(char, "bonus_table"):
		return
	
	bonus_table = objects.get_attribute(char, "bonus_table")

	# remove bonus
	for property in bonus_table:
		table_entry = bonus_table[property]
		if keyword in table_entry:
			__modify_temp_bonus(char, property, 0, keyword)
			del table_entry[keyword]
			if debug:
				print table_entry
			
	__sanity_check(char)

def remove_on_leave_party(char):
	"""Is called by object_events.on_party_left when the char leaves the player party.
	
Removes all bonuses that had remove_on_leave=True set in add_temp_bonus(es).
No callbacks are fired by this."""

	# prepare table
	bonus_table = {}
	if objects.has_attribute(char, "bonus_table"):
		bonus_table = objects.get_attribute(char, "bonus_table")

	# remove bonus
	for property in bonus_table:
		table_entry = bonus_table[property]
		keywords_to_delete = []
		for keyword in table_entry:
			if keyword not in __illegal_keywords and table_entry[keyword][REMOVE]:
				__modify_temp_bonus(char, property, 0, keyword)
				keywords_to_delete.append(keyword)
				if debug:
					print table_entry

		for keyword in keywords_to_delete:
			del table_entry[keyword]
			
	__sanity_check(char)

def reverse_temp_bonus(char, keyword, timeout=None, callback=None):
	"""Reverses the bonuses (but not the maluses) for this keyword.
	
e.g. if a char has +3 on str and -2 on con and reverse_temp_bonus is called
he has -3 on str and -2 on con.
	
Parameters:
	char     - Id of the character
	keyword  - Keyword like "increase moral effect" or "drug effect".
	timeout  - Time string like "12:00:00"
	           None if the reversed bonus has no timeout
	callback - Function to call when timeout is reached
	           If this is None remove_temp_bonus is called.
	           callback is ignored if timeout is None.
	           callback will only be called once, not for each property reversed.
"""
	# prepare table
	bonus_table = {}
	if objects.has_attribute(char, "bonus_table"):
		bonus_table = objects.get_attribute(char, "bonus_table")

	# reverse bonus
	for property in bonus_table:
		table_entry = bonus_table[property]
		if keyword in table_entry:
			value = table_entry[keyword][BONUS]
			if value > 0:
				value *= -1
				__modify_temp_bonus(char, property, value, keyword, timeout, callback, table_entry[keyword][REMOVE])
				timeout = None
				callback = None
			assert not table_entry[keyword][REVERSED] # no double reverse, reversed entries should be removed
			table_entry[keyword][REVERSED] = True
			if debug:
				print table_entry
			
	__sanity_check(char)

def is_reversed(char, keyword):
	"""Returns True if the bonus with the given keyword is reversed."""
	# prepare table
	bonus_table = {}
	if objects.has_attribute(char, "bonus_table"):
		bonus_table = objects.get_attribute(char, "bonus_table")

	# reverse bonus
	for property in bonus_table:
		table_entry = bonus_table[property]
		if keyword in table_entry:
			return table_entry[keyword][REVERSED]
			
	return False

def get_timeout (char, keyword):
	"""Returns timeout value if the bonus with the given keyword is reversed."""
	# prepare table
	bonus_table = {}
	if objects.has_attribute(char, "bonus_table"):
		bonus_table = objects.get_attribute(char, "bonus_table")

	for property in bonus_table:
		table_entry = bonus_table[property]
		if keyword in table_entry:
			return table_entry[keyword][TIMEOUT]
			
	return None

def has_keyword(char, keyword):
	"""Returns True if the char has a bonus/malus with the given keyword."""
	# prepare table
	bonus_table = {}
	if objects.has_attribute(char, "bonus_table"):
		bonus_table = objects.get_attribute(char, "bonus_table")

	# reverse bonus
	for property in bonus_table:
		table_entry = bonus_table[property]
		if keyword in table_entry:
			return True
			
	return False

def get_temp_bonus(char, property):
	"""Returns the total bonus or malus that char has on this property.

0 is returned, if the char has no bonus or malus on this property.
To get the real (unmodified) value of property do:
	object.get_attribute(char, property) - get_temp_bonus(char, property)
"""
	# prepare table
	bonus_table = {}
	if objects.has_attribute(char, "bonus_table"):
		bonus_table = objects.get_attribute(char, "bonus_table")
	
	# prepare table entry for this property
	table_entry = {}
	if property in bonus_table:
		table_entry = bonus_table[property]

	value = 0
	if "modifier" in table_entry:
		value = table_entry["modifier"]

	return value

BONUS_TYPES = {
	'reputation':globaltext.MISSION_BONUS_REPUTATION,
	'weapon_accustom_bonusleft_hand':globaltext.MISSION_BONUS_ACCUSTOM,
	'weapon_accustom_bonusright_hand':globaltext.MISSION_BONUS_ACCUSTOM,
	'weapon_accustom_bonusthrow':globaltext.MISSION_BONUS_ACCUSTOM,
	'survive':globaltext.MISSION_BONUS_SURVIVE,
	'increase_moral':globaltext.MISSION_BONUS_MORAL,
	'psychology':globaltext.MISSION_BONUS_PSYCHOLOGY,
	'drug_effects':globaltext.MISSION_BONUS_DRUGS,
	'armor':globaltext.MISSION_COMBINATION_ARMOR,
	'implantant':globaltext.MISSION_COMBINATION_IMPLANTANT
	}
	
BONUS_TYPES_ORDER=[
	'reputation',
	'weapon_accustom_bonusleft_hand',
	'weapon_accustom_bonusright_hand',
	'weapon_accustom_bonusthrow',
	'survive',
	'increase_moral',
	'psychology',
	'drug_effects',
	'armor',
	'implantant']
	

def get_boni_string(character_id, attribute):
	"""Function to generate text strings for the bonus display in the ID card."""
	
	#shortout if nothing to do
	bonus_all = get_temp_bonus(character_id, attribute)
	if bonus_all == 0:
		return ""
	
	# prepare table
	bonus_table = {}
	if objects.has_attribute(character_id, "bonus_table"):
		bonus_table = objects.get_attribute(character_id, "bonus_table")
	else:
		return ""
	
	# prepare table entry for this attribute
	if attribute in bonus_table:
		table_entry = bonus_table[attribute]
	else:
		return ""
	#print bonus_table
	#print "Attribute:", attribute,"table:", table_entry
	
	# build string
	bonus_string = ""
	for bonus_type in BONUS_TYPES_ORDER:
		#print "bonus_type",bonus_type
		if bonus_type in table_entry:
			act_bonus = table_entry[bonus_type][BONUS]
			act_bonus_string = ""
			if act_bonus > 0:
				act_bonus_string = "+" + str(act_bonus)
			else:
				act_bonus_string = str(act_bonus)
			#print "act_bonus", act_bonus
			bonus_string = bonus_string + act_bonus_string + BONUS_TYPES[bonus_type] + "\n"
	
	#print bonus_string
	return bonus_string
