# -*- coding: utf-8 -*-
__doc__ = """Module for the time class.
The time class represents the in-game time.
"""

class time:
	"""Instances of this class represent a point of time."""
	_time = 0
	def __init__(self, time, time_high = None):
		"""Class constructor.

Time is a string (see set_from_string).
"""
		assert type(time) is int or type(time) is str or type(time) is long
		if type(time) is str:
			assert time_high == None
			self.set_from_string(time)
		else:
			if not time_high:
				self._time = time
			else:
				assert type(time_high) is int or type(time_high) is str or type(time_high) is long
				self._time = time | (long(time_high) << 32)
				
	def __str__(self):
		"""Returns the time as string in format 'hh:mm:ss.msms'."""
		t = self._time
		h = t / (1000*60*60)
		m = (t / (1000*60)) % 60
		s = (t / 1000) % 60
		ms = t % 1000
		return "%.2d:%.2d:%.2d.%.4d" % (h, m, s, ms)
	def __repr__(self):
		"""Returns the time as string in format 'hh:mm:ss.msms'."""
		return "'" + str(self) + "'"
	def get_day_time(self):
		"""Gets the time at the current day.

The time can be > 24h, i.e., it can be more than one day. This function gets
the time at the current day. This is the same as time % 24h.
"""
		return self % "24:00:00"
	
	def get_small_time(self):
		"""Return time in hh:mm format"""
		t = self._time
		h = t / (1000*60*60)
		m = (t / (1000*60)) % 60
		return "%.2d:%.2d" % (h, m)		
		
	def set_day_time(self, day_time):
		"""Advances the time until the given day time is reached.

Example:
        t = time("17:00:00")
        t.set_day_time("7:00:00")
        str(t)
                --> "31:00:00.0000"
"""
		if isinstance(day_time, str): day_time = time(day_time)
		assert isinstance(day_time, time)
		assert day_time < "24:00:00"
		days = self / "24:00:00"
		rest = self % "24:00:00"
		if rest > day_time: days += 1
		self._time = (time("24:00:00") * days + day_time)._time
	def set_from_string(self, time):
		"""Sets the time from a string.

Syntax examples:
    "17:32:21.100"  --> 17 hours, 32 minutes, 21 seconds and 100 milliseconds.
    "17:32:21"      --> 17 hours, 32 minutes, 21 seconds and   0 milliseconds.
    "17:32"         --> 17 hours, 32 minutes,  0 seconds and   0 milliseconds.
    "21"            -->  0 hours,  0 minutes, 21 seconds and   0 milliseconds.
    "21.100"        -->  0 hours,  0 minutes, 21 seconds and 100 milliseconds.
"""
		assert type(time) is str

		h = 0
		m = 0
		s = 0
		ms = 0

		temp = time.split(".")
		assert 1 <= len(temp) and len(temp) <= 2
		if len(temp) == 2:
			ms = int(temp[-1])
		temp = temp[0]
		temp = temp.split(":")
		assert len(temp)
		
		if len(temp)==1 or len(temp)==3:
			s = int(temp[-1])
			temp = temp[:-1]		
		if len(temp)==2:
			m = int(temp[1])
			h = int(temp[0])
		else: assert len(temp)==0
		self._time = ms + s*1000 + m*1000*60 + h*1000*60*60

	def __eq__(self, other):
		"""time == other.

If other is not a time value its always False.
"""
		if isinstance(other, str): other = time(other)
		if not isinstance(other, time): return False
		return self._time == other._time
	def __lt__(self, other):
		"""time < other.

If other is not a time value, it's always False.
"""
		if isinstance(other, str): other = time(other)
		if not isinstance(other, time): return False
		return self._time < other._time
	def __le__(self, other):
		"""time <= other.
		
If other is not a time value, it's always False.
"""
		if isinstance(other, str): other = time(other)
		if not isinstance(other, time): return False
		return self._time <= other._time
	def __ne__(self, other):
		"""time != other.
		
If other is not a time value, it's always False.
"""
		if isinstance(other, str): other = time(other)
		if not isinstance(other, time): return False
		return self._time != other._time
	def __ge__(self, other):
		"""time >= other.
		
If other is not a time value, it's always False.
"""
		if isinstance(other, str): other = time(other)
		if not isinstance(other, time): return False
		return self._time >= other._time
	def __gt__(self, other):
		"""time > other.
		
If other is not a time value, it's always False
"""
		if isinstance(other, str): other = time(other)
		if not isinstance(other, time): return False
		return self._time > other._time
	def __add__(self, other):
		"""time + other.

Other has to be a time value.

Example:
        time("18:00:00") + time("6:00:00") == time("24:00:00").
"""
		if isinstance(other, str): other = time(other)
		assert isinstance(other, time)
		temp = self._time + other._time
		return time(temp)
	def __sub__(self, other):
		"""time - other.

Other has to be a time value.
Only positive times are legal, so if other is bigger than self, an exception is
raised.
Example:
        time("24:00:00") - time("6:00:00") == time("18:00:00").
"""
		if isinstance(other, str): other = time(other)
		assert isinstance(other, time)
		temp = self._time - other._time
		assert temp >= 0, "subtraction yielded negative result"
		return time(temp)
	def __div__(self, other):
		"""time / other.

Other has to be a time value or an integer.

Examples:
        time("24:00:00") / time("6:00:00") == 4
        time("24:00:00") / time("9:00:00") == 2
        time("24:00:00") / 12 == time("2:00:00")
"""
		if isinstance(other, str): other = time(other)
		assert type(other) is int or isinstance(other, time) or type(other) is long
		if type(other) is int or type(other) is long:
			return time(self._time / other)
		else:
			return self._time / other._time
	def __truediv__(self, other):
		"""time / other.
		
Other has to be a time value or an integer.

Examples:
        time("24:00:00") / time("6:00:00") == 4
        time("24:00:00") / time("9:00:00") == 2
        time("24:00:00") / 12 == time("2:00:00")
"""
		if isinstance(other, str): other = time(other)
		assert type(other) is int or isinstance(other, time) or type(other) is long
		if type(other) is int or type(other) is long:
			return time(self._time / other)
		else:
			return self._time / other._time
	def __floordiv__(self, other):
		"""time // other.

Other has to be a time value or an integer.

Examples:
        time("24:00:00") // time("6:00:00") == 4
        time("24:00:00") // 12 == time("2:00:00").
"""
		if isinstance(other, str): other = time(other)
		assert type(other) is int or isinstance(other, time) or type(other) is long
		if type(other) is int or type(other) is long:
			return time(self._time // other)
		else:
			return self._time // other._time
	def __mul__(self, other):
		"""time * other.

Other has to be an integer.

Examples:
        time("6:00:00") * 4 == time("24:00:00")
"""
		assert type(other) is int or type(other) is long
		temp = self._time * other
		return time(temp)
	def __mod__(self, other):
		"""time % other

Other has to be a time value.

Examples:
        time("24:00:00") % time("9:00:00") == time("6:00:00")
        time("24:00:00") % time("6:00:00") == time("0:00:00")
"""
		if isinstance(other, str): other = time(other)
		assert isinstance(other, time)
		return time(self._time % other._time)

